import {SdkConst} from "./sdk-const";
import {ClientData} from "./client-data";
import {getRandomNumber} from "./utils";
import {BefrestMessage, MsgType} from "./befrest-message";
import {BefrestConnectionMode} from "./befrest-connection-mode";

// @ts-ignore
export class BefrestInternal {
    private befrestWebSocket: WebSocket | undefined;
    private isPendingPing?: boolean;
    private pingInterval?: number;
    private pingPayload = "";
    private isRefreshingConnection?: boolean;
    private isBatchTypeReceived = false;
    private batchSize?: number;
    private pendingAcks: string[] = [];
    private pendingMessages: BefrestMessage[] = [];
    private sendingAcksInterval?: number;
    private isBefrestStart = false;
    private sendAckTimeOut?: number;
    private acks = "";
    private isBatchToSendSync = false;


    constructor() {
    }

    public start(): void {
        if (this.isBefrestStart) {
            console.log("Befrest-befrest already started");
            return;
        }
        this.isBefrestStart = true;
        console.log("Befrest-start");
        this.setupHeader();
        if (this.befrestWebSocket) {
            if (this.befrestWebSocket.readyState !== WebSocket.OPEN) {

                this.connectToServer();
            }
        } else {
            this.connectToServer();
        }
    }

    public stop(): void {
        console.log("Befrest-stop");
        if (this.befrestWebSocket) {
            this.befrestWebSocket.close();
        }
        if (this.pingInterval) {
            clearInterval(this.pingInterval);
        }
        if (this.sendingAcksInterval) {
            clearInterval(this.sendingAcksInterval);
        }
        this.isBefrestStart = false;
    }

    public refresh(): void {
        console.log("Befrest-refresh");
        if (this.befrestWebSocket == null || this.befrestWebSocket.readyState != WebSocket.OPEN) {
            this.start();
        } else {
            this.befrestWebSocket.send(SdkConst.PING_PREFIX + getRandomNumber());
            this.isPendingPing = true;
            this.setupPingTimout();
        }
    }

    private connectToServer(): void {

        console.log("Befrest-Connecting to server...");
        this.befrestWebSocket = new WebSocket(BefrestInternal.generateUrl());
        this.scheduleRestart();

        // @ts-ignore
        this.befrestWebSocket.onopen = () => {
            if (ClientData.getInstance().messageHandler) {
                ClientData.getInstance().messageHandler.onBefrestConnect();
                ClientData.getInstance().messageHandler.onConnectionChanged(BefrestConnectionMode.CONNECTED);
                this.startPinging();
            } else {
                throw Error('messageHandler can not be null');
            }
        };
        // @ts-ignore
        this.befrestWebSocket.onmessage = (message: MessageEvent) => {
            if (ClientData.getInstance().messageHandler) {
                ClientData.getInstance().messageHandler.onConnectionRefreshed();
            }
            this.cancelPingTimeout();
            if (this.checkPong(message.data)) {
                return;
            }
            const befrestMessage = new BefrestMessage(message.data);
            if (!befrestMessage.corruptionFlag) {
                console.log("Befrest-message receive : " + message.toString());
                switch (befrestMessage.getType) {
                    case MsgType.BATCH:
                        console.log("Befrest-batch message receive with size :" + message.data);
                        this.isBatchTypeReceived = true;
                        this.batchSize = 0;
                        this.pendingAcks = [];
                        if (befrestMessage.message)
                            this.prepareToReceiveMessages(befrestMessage.message);
                        break;
                    case MsgType.NORMAL:
                    case MsgType.GROUP:
                    case MsgType.TOPIC:
                        if (befrestMessage.getAckMessage)
                            if ((this.acks.length + befrestMessage.getAckMessage).length >= 2 * 1024) {
                                if (this.befrestWebSocket)
                                    this.befrestWebSocket.send(this.acks.substring(0, this.acks.length - 1));
                                this.acks = "";
                            }
                        this.acks += befrestMessage.getAckMessage;
                        clearTimeout(this.sendAckTimeOut);
                        this.acks += "&";
                        this.sendAckTimeOut = setTimeout(() => {
                            this.sendAcksV2();
                        }, 1000);


                        if (this.isBatchTypeReceived) {
                            if (this.batchSize)
                                if (this.batchSize > 0) {
                                    this.batchSize--;
                                    this.pendingMessages.push(befrestMessage);
                                }
                            if (this.batchSize === 0) {
                                this.isBatchToSendSync = true;
                                ClientData.getInstance().messageHandler.onMessagesReceived(this.pendingMessages);
                                this.pendingMessages = [];
                                this.isBatchTypeReceived = false;
                                clearTimeout(this.sendAckTimeOut);
                                this.sendAckTimeOut = setTimeout(() => {
                                    this.sendAcksV2();
                                }, 1000);
                            }
                        } else {
                            this.pendingMessages.push(befrestMessage);
                            ClientData.getInstance().messageHandler.onMessagesReceived(this.pendingMessages);
                            this.pendingMessages = [];
                        }

                        break;
                }
            }

        };
        if (this.befrestWebSocket)
            this.befrestWebSocket.onclose = (ev: any) => {
                this.isBefrestStart = false;
                console.log("Befrest-connection close with message : " + ev.toString());
                if (ClientData.getInstance().messageHandler) {
                    ClientData.getInstance().messageHandler.onConnectionChanged(BefrestConnectionMode.DISCONNECTED);
                    if (this.isRefreshingConnection) {
                        this.start();
                    }
                } else {
                    throw Error('messageHandler can not be null');
                }
            };
        if (this.befrestWebSocket)
            this.befrestWebSocket.onerror = (e: any) => {
                try {
                    this.isBefrestStart = false;
                    console.log("Befrest-error happens with message : " + e.message.toString());
                    if (ClientData.getInstance().messageHandler) {
                        ClientData.getInstance().messageHandler.onConnectionChanged(BefrestConnectionMode.DISCONNECTED);
                        if (e.message.toString().indexOf("401 Unauthorized")) {
                            ClientData.getInstance().messageHandler.onAuthorizeProblem();
                        }
                    }
                } catch (e) {
                }
            };
    }

    public commitTopics() {
        console.log("Befrest-commit topics");
        this.isRefreshingConnection = true;
        this.stop();
    }

    private startPinging() {
        if (this.befrestWebSocket == null || this.befrestWebSocket.readyState != WebSocket.OPEN || this.isPendingPing) {
            this.befrestWebSocket = undefined;
            this.start();
            return;
        }
        console.log("Befrest-start pinging with 10 sec interval");
        this.pingInterval = setInterval(() => {
            this.isPendingPing = true;
            this.pingPayload = SdkConst.PING_PREFIX + getRandomNumber();
            if (this.befrestWebSocket)
                this.befrestWebSocket.send(this.pingPayload);
            console.log("ping server with payload", this.pingPayload);
            this.setupPingTimout();
        }, 10000);
    }

    private setupHeader(): void {
        console.log("Befrest-setup header with auth token : " +
            ClientData.getInstance().authToken + " and topics : " +
            ClientData.getInstance().topics);
        SdkConst.OPTIONS = {
            headers: {
                "X-BF-AUTH": ClientData.getInstance().authToken,
                Connection: "Upgrade",
                Upgrade: "websocket",
                "Sec-WebSocket-Key": "sample",
                "Sec-WebSocket-Version": "12",
                "X-BF-TOPICS": ClientData.getInstance().topics
            }
        };
    }

    private static generateUrl(): string {
        let t = "TOPIC";
        if (ClientData.getInstance().topics.length == 0)
            t = "/TOPIC";


        return SdkConst.BASE_URL
            .replace("AID", ClientData.getInstance().aid)
            .replace("CHID", ClientData.getInstance().chid)
            .replace("AUTH", ClientData.getInstance().authToken)
            .replace(t, ClientData.getInstance().topics);
    }

    private scheduleRestart() {
        console.log("Befrest-scheduled restart after try to connect");
        setTimeout(() => {
            if (!this.befrestWebSocket || this.befrestWebSocket.readyState !== WebSocket.OPEN) {
                if (ClientData.getInstance().messageHandler) {
                    ClientData.getInstance().messageHandler.onConnectionChanged(BefrestConnectionMode.RETRY);
                    this.isBefrestStart = false;
                }
                this.start();
            }
        }, 3000);
    }

    private checkPong(data: any): boolean {
        return data.toString().toUpperCase().startsWith(SdkConst.PING_PREFIX);
    }

    private prepareToReceiveMessages(message: string) {
        this.batchSize = Number(message);
    }

    private sendAcksV2() {
        console.log("Befrest-send batch Acks " + this.acks);
        if (this.befrestWebSocket)
            if (this.befrestWebSocket.readyState === WebSocket.OPEN) {
                this.befrestWebSocket.send(this.acks.substring(0, this.acks.length - 1));
                this.acks = "";
            }
        if (this.isBatchToSendSync) {
            setTimeout(() => {
                if (this.befrestWebSocket)
                    if (this.befrestWebSocket.readyState === WebSocket.OPEN) {
                        this.befrestWebSocket.send("S");
                    }
            }, 50);
            this.isBatchToSendSync = false;
        }
    }

    private cancelPingTimeout(): void {
        this.isPendingPing = false;
    }

    private setupPingTimout(): void {
        setTimeout(() => {
            if (!this.befrestWebSocket || this.befrestWebSocket.readyState != WebSocket.OPEN || this.isPendingPing) {
                if (ClientData.getInstance().messageHandler) {
                    ClientData.getInstance().messageHandler.onConnectionChanged(BefrestConnectionMode.RETRY);
                    this.start();
                }
            }
        }, 3000);
    }
}

