import {BefrestMessageHandler} from "./befrest-message-handler";
import {BefrestInternal} from "./befrest-internal";
import {ClientData} from "./client-data";

export class Befrest {
    private static instance: Befrest;
    private befrestInternal: BefrestInternal;


    constructor() {
        this.befrestInternal = new BefrestInternal();
    }

    public static getInstance(): Befrest {
        if (!Befrest.instance) {
            Befrest.instance = new Befrest();
        }
        return Befrest.instance;
    }

    /**
     *init befrest with aid , chid  ,authToken and messageHandler
     * call this method first to use befrest
     */
    public init(aid: string, chid: string, authToken: string, messageHandler: BefrestMessageHandler): void {

        if (!aid.toString().trim().length) {
            throw Error("aid cannot be null");
        }
        if (!chid.toString().trim().length) {
            throw Error("chid cannot be null");
        }
        if (!messageHandler) {
            throw Error("messageHandler cannot be null");
        }
        if (aid !== ClientData.getInstance().aid) {
            ClientData.getInstance().aid = aid;
        }
        if (chid !== ClientData.getInstance().chid) {
            ClientData.getInstance().chid = chid;
        }
        if (authToken && authToken !== ClientData.getInstance().authToken) {
            ClientData.getInstance().authToken = authToken;
        }
        ClientData.getInstance().messageHandler = messageHandler;
    }

    /**
     *The befrest start method create connection to server to receive and send messages through websocket
     *call this method after init to start befrest
     */
    public start(): void {
        this.addTopic('x');
        this.befrestInternal.start();
    }

    /**
     * The befrest stop method disconnect from server after calling this method no message receive from befrest
     * call this method whenever you want to stop befrest
     */
    public stop(): void {
        this.befrestInternal.stop();
    }

    /**
     * The commitTopic method will be stop befrest and start again to update topic list in server
     * call this method whenever you add or remove topic(recommended)
     */
    public commitTopic(): void {
        this.befrestInternal.commitTopics();
    }

    /**
     * The refresh method send ping payload to server whenever you call to check connection status
     * and notifying you in onConnectionRefresh callback.
     */
    public refresh(): void {
        this.befrestInternal.refresh();
    }

    /**
     * The getTopicList method return a string array that show current topics that your client subscribed
     */
    public getTopicList(): string [] {
        return ClientData.getInstance().topics ? ClientData.getInstance().topics.split("-") : [];
    }

    /**
     * The addTopic method use to add specific topic
     * @param topic is a topic that you want to add
     * @return true if new topic added successfully and false if your topic has already exist
     */
    public addTopic(topic: string): boolean | null {
        if (!topic.match(/^[a-z0-9]+$/i)) {
            console.log("topic name must be alpha-numeric");
            return false;
        }
        const addedTopic = topic.trim();
        if (ClientData.getInstance().topics.length === 0) {
            ClientData.getInstance().topics = addedTopic;
            return true;
        } else if (ClientData.getInstance().topics.toString().indexOf(addedTopic) === -1) {
            ClientData.getInstance().topics += "-" + addedTopic;
            return true;
        } else {
            console.log("Befrest-topic: " + topic + " has already exist int topic list");
            return false;
        }
    }

    /**
     * The removeTopic method use to remove specific topic
     * @param topic is a topic that you want to remove
     * @return true if topic remove successfully and false if topic not exists in topic list
     */
    public removeTopic(topic: string): boolean {
        if (!topic.match(/^[a-z0-9]+$/i)) {
            console.log("topic name must be alpha-numeric");
            return false;
        }
        const toRemove = topic.trim();
        if (ClientData.getInstance().topics.length == 0) {
            console.log("Befrest-topic list is empty");
            return false;
        } else if (ClientData.getInstance().topics.toString().indexOf(toRemove) === -1) {
            console.log("Befrest-topic: " + toRemove + " not exist in topic list");
            return false;
        }
        ClientData.getInstance().topics = ClientData.getInstance().topics.replace(`${topic}-`, "");
        ClientData.getInstance().topics = ClientData.getInstance().topics.replace(`-${topic}`, "");
        return true;
    }

}